@@ -68,6 +68,8 @@
 #include "http_config.h"
 #include "http_log.h"
 
+/* Note that APACHE_SSL_EXPORT_CERTS is defined in buff.h */
+
 /*
 Useful bits for dbx (useful for Ben, at least):
 
@@ -77,6 +79,11 @@ use /u/ben/work/scuzzy-ssleay8/ssl /u/ben/work/scuzzy-ssleay8/crypto/bio
 
 #define SSL_IS_OPTIONAL		FALSE	/* Set to TRUE if you want to allow SSL to be disabled on a per-request basis (useful for subrequests) */
 
+#if SSLEAY_VERSION_NUMBER >= 0x0800
+/* Do session caching - only available with SSLeay 0.8.x */
+# define CACHE_SESSIONS		TRUE
+#endif
+
 #if SSL_IS_OPTIONAL
 # define OPTIONAL_SSL	(!r->connection->client->ssl)
 #else
@@ -88,15 +95,14 @@ use /u/ben/work/scuzzy-ssleay8/ssl /u/ben/work/scuzzy-ssleay8/crypto/bio
 /* But note, they nearly are, if anyone cares */
 #endif
 
-#if SSLEAY_VERSION_NUMBER >= 0x0800
-/* Do session caching - only available with SSLeay 0.8.x */
-# define CACHE_SESSIONS		TRUE
+#if SSLEAY_VERSION_NUMBER < 0x0900 && APACHE_SSL_EXPORT_CERTS
+# error "Can't export certificates with SSLeay before 0.9.0"
 #endif
 
 #define FREE_SESSION	TRUE	/* Don't set this FALSE! */
 
 /* define this to debug the verification stuff */
-#define DEBUG_VERIFY	FALSE
+#define DEBUG_VERIFY	TRUE
 
 #if CACHE_SESSIONS
 # include "gcache.h"
@@ -107,6 +113,10 @@ static char *s_szCacheServerPath;
 
 #endif
 
+#if APACHE_SSL_EXPORT_CERTS && !CACHE_SESSIONS
+# error "CACHE_SESSIONS must be TRUE if APACHE_SSL_EXPORT_CERTS is"
+#endif
+
 #if SSLEAY_VERSION_NUMBER < 0x900
 /* correct for a typo */
 # define SSL3_TXT_ADH_DES_192_CBC_SHA SSL3_TXT_ADH_DES_196_CBC_SHA
@@ -127,6 +137,8 @@ static char *s_szCacheServerPath;
 # define X509_verify_cert_error_string(error) X509_cert_verify_error_string(error)
 #endif
 
+#define ENV_CERT_CHAIN		"SSL_CLIENT_CERT_CHAIN_"
+
 #define UNSET	-1
 
 typedef enum
@@ -169,31 +181,44 @@ typedef struct
     table *tbRequiredCiphers;
     table *tbBannedCiphers;
     BOOL bSSLRequired;
+#if APACHE_SSL_EXPORT_CERTS
+    BOOL bExportCerts;
+#endif
     } SSLDirConfigRec;
 
 MODULE_VAR_EXPORT module apache_ssl_module;
 
 static conn_rec *s_pCurrentConnection;
+static pool *s_pPermanentPool;
 
-static const char *SSLRequireCipher(cmd_parms *cmd,SSLDirConfigRec *rec,
-				    char *cipher)
+static const char six2pr[64+1]=
+  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
+
+static void uuencoden(unsigned char *szTo,const unsigned char *szFrom,
+                     int nLength)
     {
-    ap_table_set(rec->tbRequiredCiphers,cipher,"Required");
-    return NULL;
+    const unsigned char *s;
+
+    for(s=(const unsigned char *)szFrom ; nLength > 0 ; s+=3)
+	{
+	*szTo++=six2pr[s[0] >> 2];
+	*szTo++=six2pr[(s[0] << 4 | s[1] >> 4)&0x3f];
+	if(!--nLength)
+	    break;
+	*szTo++=six2pr[(s[1] << 2 | s[2] >> 6)&0x3f];
+	if(!--nLength)
+	    break;
+	*szTo++=six2pr[s[2]&0x3f];
+	if(!--nLength)
+	    break;
+	}
+    *szTo++='\0';
     }
 
-static const char *SSLBanCipher(cmd_parms *cmd,SSLDirConfigRec *rec,
-				char *cipher)
+static void uuencode(char *szTo,const char *szFrom)
     {
-    ap_table_set(rec->tbBannedCiphers,cipher,"Banned");
-    return NULL;
-    }
-
-static const char *SSLRequireSSL(cmd_parms *cmd,SSLDirConfigRec *rec,
-				 char *cipher)
-    {
-    rec->bSSLRequired=TRUE;
-    return NULL;
+    uuencoden((unsigned char *)szTo,(const unsigned char *)szFrom,
+	      strlen(szFrom));
     }
 
 static int SSLCheckCipher(request_rec *r)
@@ -247,32 +272,46 @@ static void StrUpper(char *s)
 	*s=toupper(*s);
     }
 
-static void ExpandCert(table *pEnv,char *szPrefix,char *szCert)
+static void ExpandCert(pool *p,table *pEnv,char *szPrefix,char *szCert)
     {
-    /* Use ap_pstrintf when we get to 1.3 */
-    char var[HUGE_STRING_LEN];
     char buf[HUGE_STRING_LEN];
-    char *s;
+    char *s,*t;
     /* Expand a X509_oneline entry into it's base components and register
        them as environment variables. Needed if you want to pass certificate
        information to CGI's. The naming convention SHOULD be fairly compatible
        with CGI's written for stronghold's certificate info  - Q */
     /* FIXME - strtok() and strcspn() may cause problems on some systems - Q */
 
-    sprintf(var,"%sDN",szPrefix);
-    ap_table_set(pEnv,var,szCert);
+    t=ap_psprintf(p,"%sDN",szPrefix);
+    ap_table_setn(pEnv,t,ap_pstrdup(p,szCert));
     
-    strcpy(buf,szCert);
+    ap_cpystrn(buf,szCert,sizeof buf);
     for(s=strtok(buf,"/") ; s != NULL ; s=strtok(NULL,"/"))
 	{
 	int n=strcspn(s,"=");
 	s[n]='\0';
 	StrUpper(s);
-	sprintf(var,"%s%s",szPrefix,s);
-	ap_table_set(pEnv,var,s+n+1);
+	t=ap_pstrcat(p,szPrefix,s,NULL);
+	ap_table_setn(pEnv,t,ap_pstrdup(p,s+n+1));
 	}
     }
 
+#if APACHE_SSL_EXPORT_CERTS
+/* szEnv must have been pool allocated, or be static */
+static void AddCertToEnv(pool *p,table *e,char *szEnv,X509 *xs)
+    {
+    int n;
+    unsigned char *s,*d;
+
+    n=i2d_X509(xs,NULL);
+    d=s=ap_palloc(p,n);
+    i2d_X509(xs,&d);
+    d=ap_palloc(p,(n*4)/3+2);
+    uuencoden(d,s,n);
+    ap_table_setn(e,szEnv,(char *)d);
+    }
+#endif
+
 static struct
     {
     char *szName;
@@ -327,13 +366,20 @@ static int SSLFixups(request_rec *r)
     table *e=r->subprocess_env;
     int keysize=0;
     int secretkeysize=0;
-    char buf[MAX_STRING_LEN];
+    char *t;
     char *cipher;
     int n;
     X509 *xs;
+#if APACHE_SSL_EXPORT_CERTS && !APACHE_SSL_KEEP_CERTS
+    STACK *pCertStack;
+#endif
 
     SSLConfigRec *pConfig=(SSLConfigRec *)
       ap_get_module_config(r->server->module_config,&apache_ssl_module);
+#if APACHE_SSL_EXPORT_CERTS
+    SSLDirConfigRec *pDirConfig=(SSLDirConfigRec *)
+      ap_get_module_config(r->per_dir_config,&apache_ssl_module);
+#endif
 
     /* Check to see if SSL is on */
     if(pConfig->bDisabled || OPTIONAL_SSL)
@@ -355,25 +401,47 @@ static int SSLFixups(request_rec *r)
 	    break;
 	    }
 
-    sprintf(buf,"%d",keysize);
-    ap_table_set(e,"HTTPS_KEYSIZE",buf);
+    t=ap_psprintf(r->pool,"%d",keysize);
+    ap_table_setn(e,"HTTPS_KEYSIZE",t);
 
-    sprintf(buf,"%d",secretkeysize);
-    ap_table_set(e,"HTTPS_SECRETKEYSIZE",buf);
+    t=ap_psprintf(r->pool,"%d",secretkeysize);
+    ap_table_setn(e,"HTTPS_SECRETKEYSIZE",t);
 
     if(r->connection->client->szClientX509)
 	{
-	ExpandCert(e,"SSL_CLIENT_",r->connection->client->szClientX509);
+	ExpandCert(r->pool,e,"SSL_CLIENT_",
+		   r->connection->client->szClientX509);
 	xs=SSL_get_peer_certificate(r->connection->client->ssl);
-	ExpandCert(e,"SSL_CLIENT_I",X509_NAME_oneline(X509_get_issuer_name(xs),
-						      NULL,0));
+	ExpandCert(r->pool,e,"SSL_CLIENT_I",
+		   X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
+#if APACHE_SSL_EXPORT_CERTS
+	if(pDirConfig->bExportCerts)
+	    AddCertToEnv(r->pool,e,"SSL_CLIENT_CERT",xs);
+#endif
 	}
 
     xs=SSL_get_certificate(r->connection->client->ssl);
-    ExpandCert(e,"SSL_SERVER_",X509_NAME_oneline(X509_get_subject_name(xs),
-						 NULL,0));
-    ExpandCert(e,"SSL_SERVER_I",X509_NAME_oneline(X509_get_issuer_name(xs),
-						  NULL,0));
+    ExpandCert(r->pool,e,"SSL_SERVER_",
+	       X509_NAME_oneline(X509_get_subject_name(xs),NULL,0));
+    ExpandCert(r->pool,e,"SSL_SERVER_I",
+	       X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
+
+#if APACHE_SSL_EXPORT_CERTS
+# if !APACHE_SSL_KEEP_CERTS
+    pCertStack=SSL_get_peer_cert_chain(r->connection->client->ssl);
+    /* XXX: When this works, check whether we need to start at 1, or add an offset */
+    for(n=1 ; pCertStack && n < sk_num(pCertStack) ; ++n)
+	{
+	t=ap_psprintf(r->pool,ENV_CERT_CHAIN "%d",n);
+	AddCertToEnv(r->connection->client->pool,e,t,(X509 *)sk_value(pCertStack,n));
+	}
+# else
+    if(r->connection->client->tableCertChain && pDirConfig->bExportCerts)
+	r->subprocess_env=
+	  ap_overlay_tables(r->pool,r->subprocess_env,
+			    r->connection->client->tableCertChain);
+# endif
+#endif
 
     return DECLINED;
     }
@@ -460,6 +528,9 @@ static void *CreateSSLDirConfig(pool *p,char *dummy)
     rec->tbRequiredCiphers=ap_make_table(p,4);
     rec->tbBannedCiphers=ap_make_table(p,4);
     rec->bSSLRequired=FALSE;
+#if APACHE_SSL_EXPORT_CERTS
+    rec->bExportCerts=FALSE;
+#endif
 
     return rec;
     }
@@ -476,13 +547,14 @@ static void *MergeSSLDirConfig(pool *p,void *basev,void *addv)
     new->tbBannedCiphers=ap_overlay_tables(p,add->tbBannedCiphers,
 					   base->tbBannedCiphers);
     new->bSSLRequired=base->bSSLRequired || add->bSSLRequired;
+#if APACHE_SSL_EXPORT_CERTS
+    new->bExportCerts=base->bExportCerts || add->bExportCerts;
+#endif
     return new;
     }
 
 static void InitSSL(server_rec *s)
     {
-    char *CAfile=NULL,*CApath=NULL;
-
 #if CACHE_SESSIONS
     if(!s_szCacheServerPort)
 	{
@@ -538,6 +610,7 @@ static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
 #endif
     SSLConfigRec *pConfig=ap_get_module_config(s_pCurrentConnection->server->module_config,&apache_ssl_module);
     server_rec *pServer=s_pCurrentConnection->server;
+    pool *p=s_pCurrentConnection->client->pool;
 
 #if DEBUG_VERIFY
       {
@@ -560,6 +633,19 @@ static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
       free(s);
       }
 #endif
+    
+#if APACHE_SSL_KEEP_CERTS
+    if(depth)
+	{
+	char *t;
+
+	if(!s_pCurrentConnection->client->tableCertChain)
+	    s_pCurrentConnection->client->tableCertChain=ap_make_table(p,5);
+
+	t=ap_psprintf(p,ENV_CERT_CHAIN "%d",depth);
+	AddCertToEnv(p,s_pCurrentConnection->client->tableCertChain,t,xs);
+	}
+#endif
 
     if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE)
 	{
@@ -618,6 +704,8 @@ static int VerifyFlags(SSLConfigRec *pConfig)
     case VERIFY_OPTIONAL_NO_CA:
 	nVerify|=SSL_VERIFY_PEER;
 	break;
+
+    default: break; /* avoid warnings from gcc */
 	}
     return nVerify;
     }
@@ -626,6 +714,112 @@ static int VerifyFlags(SSLConfigRec *pConfig)
 
 #define MAX_SESSION_DER	100000
 
+/*
+  If *pbLocal is TRUE, only do a local search. Also, set *pbLocal if the key
+  was found locally.
+  */
+static uchar *CacheFind(uchar *aucKey,int nKey,int *pnData,BOOL *pbLocal)
+    {
+    Cache *p;
+    uchar *t;
+
+    p=LocalCacheFind(aucKey,nKey);
+
+    if(!p)
+	{
+	uchar *t2;
+	time_t tExpiresAt;
+
+	if(*pbLocal)
+	    return NULL;
+
+	t=GlobalCacheGet(aucKey,nKey,pnData,&tExpiresAt);
+	if(!t)
+	    return NULL;
+
+	t2=malloc(nKey);
+	memcpy(t2,aucKey,nKey);
+	p=LocalCacheAdd(t2,nKey,tExpiresAt);
+	p->aucData=t;
+	p->nData=*pnData;
+	}
+    else
+	{
+	t=p->aucData;
+	*pnData=p->nData;
+	*pbLocal=TRUE;
+	}
+    return t;
+    }
+
+static uchar *CacheFindS(char *aucKey,int nKey,int *pnData,BOOL *pbLocal)
+    { return CacheFind((uchar *)aucKey,nKey,pnData,pbLocal); }
+
+static uchar *MakeCertKey(Cache *pCache,const char *szKey,int *pnLength)
+    {
+    uchar *buf=malloc(*pnLength=pCache->nKey+strlen(szKey));
+    memcpy(buf,pCache->aucKey,pCache->nKey);
+    memcpy(buf+pCache->nKey,szKey,strlen(szKey));
+
+    return buf;
+    }
+
+static int SendCertToServer(void *extra,const char *szKey,const char *szValue)
+    {
+    Cache *pCache=extra;
+    int nKey;
+    uchar *aucKey;
+    Cache *pCertCache;
+
+    aucKey=MakeCertKey(pCache,szKey,&nKey);
+    pCertCache=LocalCacheAdd(aucKey,nKey,pCache->tExpiresAt);
+    pCertCache->aucData=(uchar *)strdup(szValue);
+    pCertCache->nData=strlen(szValue)+1;
+    GlobalCacheAdd(pCertCache->aucKey,pCertCache->nKey,pCertCache->aucData,
+		   pCertCache->nData,pCertCache->tExpiresAt);
+    return 1;
+    }
+
+#if APACHE_SSL_EXPORT_CERTS
+static void SendCertChainToServer(Cache *pCache)
+    {
+    if(!s_pCurrentConnection->client->tableCertChain)
+	return;
+    ap_table_do(SendCertToServer,pCache,
+		s_pCurrentConnection->client->tableCertChain,NULL);
+    }
+
+static void GetCertChainFromServer(uchar *aucSessionID,int nLength,
+				   BOOL bLocal)
+    {
+    int n;
+    char buf[MAX_SESSION_DER];
+
+    ap_assert(nLength < MAX_SESSION_DER);
+    memcpy(buf,aucSessionID,nLength);
+    for(n=1 ; ; ++n)
+	{
+	uchar *t;
+	int nData;
+	int l;
+
+	ap_snprintf(buf+nLength,sizeof buf-nLength,ENV_CERT_CHAIN "%d",n);
+	l=nLength+strlen(buf+nLength);
+
+	t=CacheFindS(buf,l,&nData,&bLocal);
+	if(!t)
+	    break;
+
+	if(!s_pCurrentConnection->client->tableCertChain)
+	    s_pCurrentConnection->client->tableCertChain=
+	      ap_make_table(s_pCurrentConnection->client->pool,5);
+
+	ap_table_set(s_pCurrentConnection->client->tableCertChain,
+		     buf+nLength,(char *)t);
+	}
+    }
+#endif
+
 static void SendSessionToServer(SSL_SESSION *pSession)
     {
     uchar buf[MAX_SESSION_DER];
@@ -656,39 +850,31 @@ static void SendSessionToServer(SSL_SESSION *pSession)
     p->nData=nLength;
 
     GlobalCacheAdd(p->aucKey,p->nKey,p->aucData,p->nData,tExpiresAt);
+#if APACHE_SSL_EXPORT_CERTS
+    SendCertChainToServer(p);
+#endif
     }
 
 static SSL_SESSION *GetSessionFromServer(uchar *aucSessionID,int nLength)
     {
-    int n;
     SSL_SESSION *pSession;
-    Cache *p;
-    uchar *t,*t2;
+    uchar *t;
     int nData;
+    BOOL bLocal=FALSE;
 
-    p=LocalCacheFind(aucSessionID,nLength);
-    if(!p)
-	{
-	time_t tExpiresAt;
+#if APACHE_SSL_EXPORT_CERTS
+    GetCertChainFromServer(aucSessionID,nLength,bLocal);
+#endif
+
+    t=CacheFind(aucSessionID,nLength,&nData,&bLocal);
+    if(!t)
+	return NULL;
 
-	t=GlobalCacheGet(aucSessionID,nLength,&nData,&tExpiresAt);
-	if(!t)
-	    return NULL;
-	t2=malloc(nLength);
-	memcpy(t2,aucSessionID,nLength);
-	p=LocalCacheAdd(t2,nLength,tExpiresAt);
-	p->aucData=t;
-	p->nData=nData;
-	}
-    else
-	{
-	t=p->aucData;
-	nData=p->nData;
-	}
     pSession=d2i_SSL_SESSION(NULL,&t,nData);
     if(!pSession)
 	SSLLogErrorToFile(stderr);
     assert(pSession);
+
     return pSession;
     }
 
@@ -709,7 +895,7 @@ static int NewSessionCacheCallback(SSL *pSSL,SSL_SESSION *pNew)
 
 static SSL_SESSION *GetSessionCacheCallback(SSL *pSSL,unsigned char *aucSessionID,int nLength,int *pCopy)
     {
-    SSL_SESSION *pSession,*p2;
+    SSL_SESSION *pSession;
 
 #if DEBUG_SESSIONS
     int n;
@@ -727,14 +913,64 @@ static SSL_SESSION *GetSessionCacheCallback(SSL *pSSL,unsigned char *aucSessionI
 
 #endif /* CACHE_SESSIONS */
 
+static void GetPrivateKey(server_rec *s,SSLConfigRec *pConfig,
+			  const char *szKeyFile)
+    {
+    static table *s_pKeyTable;
+    RSA *prsaKey;
+    FILE *f;
+    int nCount;
+
+    if(!s_pKeyTable)
+	{
+	s_pKeyTable=ap_make_table(s_pPermanentPool,10);
+	assert(s_pKeyTable);
+	}
+    prsaKey=(RSA *)ap_table_get(s_pKeyTable,szKeyFile);
+    if(prsaKey)
+	{
+	pConfig->prsaKey=prsaKey;
+	return;
+	}
+
+    fprintf(stderr,"Reading key for server %s:%d\n",s->server_hostname,
+	    s->port);
+
+    f=fopen(szKeyFile,"r");
+    if(!f)
+	{
+	ap_log_error(APLOG_MARK,APLOG_CRIT,s,"Can't open key file %s",
+		     szKeyFile);
+	exit(1);
+	}
+
+    pConfig->prsaKey=RSA_new();
+    for(nCount=0 ; !PEM_read_RSAPrivateKey(f,&pConfig->prsaKey,NULL) ; ++nCount)
+	{
+	if(nCount < 3 && ERR_peek_error() == ERR_PACK(ERR_LIB_EVP,
+						      EVP_F_EVP_DECRYPTFINAL,
+						      EVP_R_BAD_DECRYPT))
+	    {
+	    fprintf(stderr,"Bad passphrase - try again\n");
+	    rewind(f);
+	    continue;
+	    }
+	ap_log_error(APLOG_MARK,APLOG_CRIT,s,
+		     "Error reading private key file %s: ",szKeyFile);
+	SSLLogError(s,APLOG_CRIT,NULL);
+	exit(1);
+	}
+    fclose(f);
+
+    ap_table_setn(s_pKeyTable,ap_pstrdup(s_pPermanentPool,szKeyFile),
+		  (const char *)pConfig->prsaKey);
+    }
+
 static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
     {
-    char buf[1024],prompt[1024];
-    int n;
     FILE *f;
     char szPath[MAX_STRING_LEN];
 
-
     if(pConfig->px509Certificate)
 	{
 	ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_CRIT,s,
@@ -743,9 +979,6 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 	exit(1);
 	}
 
-    fprintf(stderr,"Reading certificate and key for server %s:%d\n",
-	    s->server_hostname,s->port);
-
 #if SSLEAY_VERSION_NUMBER < 0x0800
     pConfig->pSSLCtx=SSL_CTX_new();
 #else
@@ -756,6 +989,12 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 #if CACHE_SESSIONS
     SSL_CTX_sess_set_new_cb(pConfig->pSSLCtx,NewSessionCacheCallback);
     SSL_CTX_sess_set_get_cb(pConfig->pSSLCtx,GetSessionCacheCallback);
+# if APACHE_SSL_EXPORT_CERTS
+    SSL_CTX_set_session_cache_mode(pConfig->pSSLCtx,
+				   SSL_CTX_get_session_cache_mode(pConfig
+								  ->pSSLCtx)
+				   |SSL_SESS_CACHE_NO_INTERNAL_LOOKUP);
+# endif
 #endif
 
     if(pConfig->szReqCiphers != NULL)
@@ -781,7 +1020,7 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
     f=fopen(pConfig->szCertificateFile,"r");
     if(!f)
 	{
-	sprintf(szPath,"%s/%s",X509_get_default_cert_dir(),
+	ap_snprintf(szPath,sizeof szPath,"%s/%s",X509_get_default_cert_dir(),
 		pConfig->szCertificateFile);
 	f=fopen(szPath,"r");
 	if(!f)
@@ -810,26 +1049,11 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 	if(*pConfig->szKeyFile == '/')
 	    strcpy(szPath,pConfig->szKeyFile);
 	else
-	    sprintf(szPath,"%s/private/%s",X509_get_default_cert_area(),
-		    pConfig->szKeyFile);
+	    ap_snprintf(szPath,sizeof szPath,"%s/private/%s",
+			X509_get_default_cert_area(),pConfig->szKeyFile);
 
 /* Otherwise the path already contains the name of the certificate file */
-    f=fopen(szPath,"r");
-    if(!f)
-	{
-	ap_log_error(APLOG_MARK,APLOG_CRIT,s,"Can't open key file %s",szPath);
-	exit(1);
-	}
-
-    pConfig->prsaKey=RSA_new();
-    if(!PEM_read_RSAPrivateKey(f,&pConfig->prsaKey,NULL))
-	{
-	ap_log_error(APLOG_MARK,APLOG_CRIT,s,
-		     "Error reading private key file %s: ",szPath);
-	SSLLogError(s,APLOG_CRIT,NULL);
-	exit(1);
-	}
-    fclose(f);
+    GetPrivateKey(s,pConfig,szPath);
 
 #if SSLEAY_VERSION_NUMBER >= 0x0800
     SSL_CTX_set_tmp_rsa_callback(pConfig->pSSLCtx,TmpRSACallback);
@@ -845,15 +1069,15 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 
 #if CACHE_SESSIONS
 static server_rec *s_pServer;
+extern uid_t ap_user_id;
 
 static int CacheLauncher(void *cmd,child_info *pInfo)
     {
     char a1[1024];
-    extern uid_t ap_user_id;
 
     ap_error_log2stderr(s_pServer);
     dup2(STDERR_FILENO,STDOUT_FILENO);
-    sprintf(a1,"%d",ap_user_id);
+    ap_snprintf(a1,sizeof a1,"%d",ap_user_id);
     ap_cleanup_for_exec();
     execl(cmd,"gcache",a1,s_szCacheServerPort,NULL);
     perror(cmd);
@@ -874,11 +1098,9 @@ static void LaunchCache(server_rec *s,pool *pPool)
 
 static void InitSSLServer(server_rec *s,pool *p)
     {
-    char *szLogFile;
-    static BOOL bFirstTime=TRUE;
-    static int nStashStdin,nStashStderr,nSaveStdin,nSaveStderr;
     int n;
     server_rec *pMainServer=s;
+    int nSaveStderr;
 
     /* If you were going to make Apache-SSL add a version string by the new
        method, here's where to do it. However, it doesn't seem right to me to
@@ -889,35 +1111,15 @@ static void InitSSLServer(server_rec *s,pool *p)
 
        */
 
-#ifdef SHARED_MODULE
-    /* Allegedly this is only run once when included as a DSO (though the value
-     of doing that is questionable, since the rest of the code doesn't work
-     without this module).
-     */
-    bFirstTime=FALSE;
-#endif
-
-    if(bFirstTime)
+    if(!s_pPermanentPool)
 	{
-	bFirstTime=FALSE;
-	nStashStdin=dup(0);
-	assert(nStashStdin > 0);
-
-	nStashStderr=dup(1);	/* Yes, I really do mean 1 (2 is already a logfile) */
-	assert(nStashStderr > 1);
-	return;
+	s_pPermanentPool=ap_make_sub_pool(NULL);
+	assert(s_pPermanentPool);
 	}
 
-    nSaveStdin=dup(0);
-    assert(nSaveStdin > 0);
-    close(0);
-    n=dup(nStashStdin);
-    assert(n == 0);
-
     nSaveStderr=dup(2);
-    assert(nSaveStderr > 0);
     close(2);
-    n=dup(nStashStderr);
+    n=dup(1);
     assert(n == 2);
 
     InitSSL(s);
@@ -976,163 +1178,9 @@ static void InitSSLServer(server_rec *s,pool *p)
 #if CACHE_SESSIONS
     LaunchCache(pMainServer,p);
 #endif
-    close(0);
-    n=dup(nSaveStdin);
-    assert(n == 0);
-    close(nSaveStdin);
-
     close(2);
     n=dup(nSaveStderr);
     assert(n == 2);
-    close(nSaveStderr);
-    }
-
-#if CACHE_SESSIONS
-static const char *SSLCacheServerPort(cmd_parms *cmd,char *struct_ptr,
-				      char *arg)
-    {
-    if(strspn(arg,"0123456789") == strlen(arg))
-	s_szCacheServerPort=arg;
-    else
-	s_szCacheServerPort=ap_server_root_relative(cmd->pool,arg);
-    InitGlobalCache(s_szCacheServerPort);
-
-    return NULL;
-    }
-
-static const char *SSLCacheServerPath(cmd_parms *cmd,char *struct_ptr,
-				      char *arg)
-    {
-    s_szCacheServerPath=ap_server_root_relative(cmd->pool,arg);
-
-    return NULL;
-    }
-#endif
-    
-
-static const char *SSLLogFile(cmd_parms *cmd,char *struct_ptr,char *arg)
-    {
-    return "The SSLLogFile directive is no longer used. All logging information now goes to the error log.";
-    }
-
-static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,char *arg)
-    {
-    /* This one should be pretty generic... */
-    char *pConfig=ap_get_module_config(cmd->server->module_config,
-				       &apache_ssl_module);
-  
-    int offset=(int)cmd->info; 
-    *(char **)(pConfig+offset)=arg;
-    return NULL;
-    }
-
-static const char *set_server_int_slot (cmd_parms *cmd,char *struct_ptr,char *arg)
-    {
-  /* This one should be pretty generic... */
-    char *pConfig=ap_get_module_config(cmd->server->module_config,
-				       &apache_ssl_module);
-  
-    int offset=(int)cmd->info; 
-    *(int *)(pConfig+offset)=atoi(arg);
-    return NULL;
-    }
-
-static const char *set_server_bool_slot (cmd_parms *cmd,char *struct_ptr)
-    {
-  /* This one should be pretty generic... */
-    char *pConfig=ap_get_module_config(cmd->server->module_config,
-				       &apache_ssl_module);
-  
-    int offset=(int)cmd->info; 
-    *(BOOL *)(pConfig+offset)=TRUE;
-    return NULL;
-    }
-
-static const char *set_server_bool_slot_false (cmd_parms *cmd,char *struct_ptr)
-    {
-  /* This one should be pretty generic... */
-    char *pConfig=ap_get_module_config(cmd->server->module_config,
-				       &apache_ssl_module);
-  
-    int offset=(int)cmd->info; 
-    *(BOOL *)(pConfig+offset)=FALSE;
-    return NULL;
-    }
-
-command_rec ssl_cmds[]=
-    {
-    { "SSLDisable",set_server_bool_slot,
-      (void *)XtOffsetOf(SSLConfigRec,bDisabled),RSRC_CONF,NO_ARGS,
-      "Disable SSL" },
-    { "SSLEnable",set_server_bool_slot_false,
-      (void *)XtOffsetOf(SSLConfigRec,bDisabled),RSRC_CONF,NO_ARGS,
-      "Enable SSL" },
-    { "SSLCertificateFile",set_server_string_slot,
-      (void *)XtOffsetOf(SSLConfigRec,szCertificateFile),RSRC_CONF,TAKE1,
-      "PEM certificate file" },
-    { "SSLCertificateKeyFile",set_server_string_slot,
-      (void *)XtOffsetOf(SSLConfigRec,szKeyFile),RSRC_CONF,TAKE1,
-      "Certificate private key file (assumed to be SSLCertificateFile if absent)" },
-    { "SSLCACertificatePath",set_server_string_slot,
-      (void *)XtOffsetOf(SSLConfigRec,szCACertificatePath),RSRC_CONF,TAKE1,
-      "CA Certificate path (taken from SSL_CERT_DIR if absent)" },
-    { "SSLCACertificateFile",set_server_string_slot,
-      (void *)XtOffsetOf(SSLConfigRec,szCACertificateFile),RSRC_CONF,TAKE1,
-      "CA Certificate file (taken from SSL_CERT_FILE if absent)" },
-    { "SSLVerifyDepth",set_server_int_slot,
-      (void *)XtOffsetOf(SSLConfigRec,nVerifyDepth),RSRC_CONF,TAKE1,
-      "Verify depth (default 0)" },
-    { "SSLVerifyClient",set_server_int_slot,
-      (void *)XtOffsetOf(SSLConfigRec,nVerifyClient),RSRC_CONF,TAKE1,
-      "Verify client (0=no,1=optional,2=required" },
-    { "SSLFakeBasicAuth",set_server_bool_slot,
-      (void *)XtOffsetOf(SSLConfigRec,bFakeBasicAuth),RSRC_CONF,NO_ARGS,
-      "Translate client X509 into a user name" },
-    { "SSLRequiredCiphers",set_server_string_slot,
-      (void *)XtOffsetOf(SSLConfigRec,szReqCiphers),RSRC_CONF,TAKE1,
-      "Colon-delimited list of required ciphers" },
-    /* Per Directory */
-    { "SSLRequireCipher",SSLRequireCipher,NULL,OR_FILEINFO,ITERATE,
-      "Add a cipher to the per directory list of required ciphers" },
-    { "SSLBanCipher",SSLBanCipher,NULL,OR_FILEINFO,ITERATE,
-      "Add a cipher to the per directory list of banned ciphers" },
-    { "SSLRequireSSL",SSLRequireSSL,NULL,OR_FILEINFO,NO_ARGS,
-      "Require SSL (use to protect directories that should not be served unencrypted)" },
-#if CACHE_SESSIONS
-    { "SSLCacheServerPort",SSLCacheServerPort,NULL,RSRC_CONF,TAKE1,
-      "Cache server TCP port number" },
-    { "SSLSessionCacheTimeout",set_server_int_slot,
-      (void *)XtOffsetOf(SSLConfigRec,tSessionCacheTimeout),RSRC_CONF,TAKE1,
-      "Lifetime in seconds of session cache objects" },
-    { "SSLCacheServerPath",SSLCacheServerPath,NULL,RSRC_CONF,TAKE1,
-      "Path of the global cache server" },
-#endif
-    { "SSLLogFile",SSLLogFile,NULL,RSRC_CONF,TAKE1,
-      "Defunct SSL log file - DO NOT USE" },
-    { NULL },
-    };
-
-static const char six2pr[64+1]=
-  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
-
-void uuencode(char *szTo,const char *szFrom)
-    {
-    const unsigned char *s;
-
-    for(s=(const unsigned char *)szFrom ; *s ; s+=3)
-	{
-	*szTo++=six2pr[s[0] >> 2];
-	*szTo++=six2pr[(s[0] << 4 | s[1] >> 4)&0x3f];
-	if(!s[0])
-	    break;
-	*szTo++=six2pr[(s[1] << 2 | s[2] >> 6)&0x3f];
-	if(!s[1])
-	    break;
-	*szTo++=six2pr[s[2]&0x3f];
-	if(!s[2])
-	    break;
-	}
-    *szTo++='\0';
     }
 
 /*
@@ -1151,6 +1199,7 @@ static int FakeBasicAuth(request_rec *r)
     SSLConfigRec *pConfig=ap_get_module_config(r->server->module_config,
 					       &apache_ssl_module);
     char b1[MAX_STRING_LEN],b2[MAX_STRING_LEN];
+    char *t;
   
     if(OPTIONAL_SSL)
 	return DECLINED;
@@ -1170,10 +1219,11 @@ static int FakeBasicAuth(request_rec *r)
        We need, therefore, to provide a password. This password can be matched by
        adding the string "xxj31ZMTZzkVA" as the password in the user file.
 */
-    sprintf(b1,"%s:password",r->connection->client->szClientX509);
+    ap_snprintf(b1,sizeof b1,"%s:password",
+		r->connection->client->szClientX509);
     uuencode(b2,b1);
-    sprintf(b1,"Basic %s",b2);
-    ap_table_set(r->headers_in,"Authorization",b1);
+    t=ap_psprintf(r->pool,"Basic %s",b2);
+    ap_table_setn(r->headers_in,"Authorization",t);
     return DECLINED;
     }
 
@@ -1183,7 +1233,6 @@ static BOOL ApacheSSLSetCertStuff(conn_rec *conn)
 					       &apache_ssl_module);
     SSL *con=conn->client->ssl;
     char *cert_file=pConfig->szCertificateFile;
-    char *key_file=pConfig->szKeyFile;
 
 /*  PEM_set_getkey_callback(ApacheSSLKeyCallback); */
     if(cert_file != NULL)
@@ -1204,9 +1253,11 @@ static BOOL ApacheSSLSetCertStuff(conn_rec *conn)
 
 static void ApacheSSLSetupVerify(conn_rec *conn)
     {
+#if SSLEAY_VERSION_NUMBER < 0x0800
     SSLConfigRec *pConfig=ap_get_module_config(conn->server->module_config,
 					       &apache_ssl_module);
     int nVerify=VerifyFlags(pConfig);
+#endif
 
     conn->client->szClientX509=NULL;
 
@@ -1246,9 +1297,7 @@ int ApacheSSLSetupConnection(conn_rec * conn)
     server_rec *srvr=conn->server;
     BUFF *fb=conn->client;
     SSLConfigRec *pConfig=ap_get_module_config(srvr->module_config,
-					      &apache_ssl_module);
-    char *cert_file=pConfig->szCertificateFile;
-    char *key_file=pConfig->szKeyFile;
+					       &apache_ssl_module);
 
     if(pConfig->bDisabled)
 	{
@@ -1258,6 +1307,9 @@ int ApacheSSLSetupConnection(conn_rec * conn)
     
     s_pCurrentConnection=conn;
 
+#if APACHE_SSL_KEEP_CERTS
+    fb->tableCertChain=NULL;
+#endif
     fb->ssl=SSL_new(pConfig->pSSLCtx);
     SSL_set_fd(fb->ssl,fb->fd);
 #if FREE_SESSION
@@ -1313,6 +1365,164 @@ int ApacheSSLSetupConnection(conn_rec * conn)
     return TRUE;
     }
 
+static const char *SSLLogFile(cmd_parms *cmd,char *struct_ptr,char *arg)
+    {
+    return "The SSLLogFile directive is no longer used. All logging information now goes to the error log.";
+    }
+
+static const char *SSLRequireCipher(cmd_parms *cmd,SSLDirConfigRec *rec,
+				    char *cipher)
+    {
+    ap_table_set(rec->tbRequiredCiphers,cipher,"Required");
+    return NULL;
+    }
+
+static const char *SSLBanCipher(cmd_parms *cmd,SSLDirConfigRec *rec,
+				char *cipher)
+    {
+    ap_table_set(rec->tbBannedCiphers,cipher,"Banned");
+    return NULL;
+    }
+
+static const char *SSLRequireSSL(cmd_parms *cmd,SSLDirConfigRec *rec,
+				 char *cipher)
+    {
+    rec->bSSLRequired=TRUE;
+    return NULL;
+    }
+
+#if APACHE_SSL_EXPORT_CERTS
+static const char *SSLExportCerts(cmd_parms *cmd,SSLDirConfigRec *rec,
+				  char *cipher)
+    {
+    rec->bExportCerts=TRUE;
+    return NULL;
+    }
+#endif
+
+#if CACHE_SESSIONS
+static const char *SSLCacheServerPort(cmd_parms *cmd,char *struct_ptr,
+				      char *arg)
+    {
+    if(strspn(arg,"0123456789") == strlen(arg))
+	s_szCacheServerPort=arg;
+    else
+	s_szCacheServerPort=ap_server_root_relative(cmd->pool,arg);
+    InitGlobalCache(s_szCacheServerPort);
+
+    return NULL;
+    }
+
+static const char *SSLCacheServerPath(cmd_parms *cmd,char *struct_ptr,
+				      char *arg)
+    {
+    s_szCacheServerPath=ap_server_root_relative(cmd->pool,arg);
+
+    return NULL;
+    }
+#endif
+
+static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,char *arg)
+    {
+    /* This one should be pretty generic... */
+    char *pConfig=ap_get_module_config(cmd->server->module_config,
+				       &apache_ssl_module);
+  
+    int offset=(int)(long)cmd->info; 
+    *(char **)(pConfig+offset)=arg;
+    return NULL;
+    }
+
+static const char *set_server_int_slot (cmd_parms *cmd,char *struct_ptr,char *arg)
+    {
+  /* This one should be pretty generic... */
+    char *pConfig=ap_get_module_config(cmd->server->module_config,
+				       &apache_ssl_module);
+  
+    int offset=(int)(long)cmd->info; 
+    *(int *)(pConfig+offset)=atoi(arg);
+    return NULL;
+    }
+
+static const char *set_server_bool_slot (cmd_parms *cmd,char *struct_ptr)
+    {
+  /* This one should be pretty generic... */
+    char *pConfig=ap_get_module_config(cmd->server->module_config,
+				       &apache_ssl_module);
+  
+    int offset=(int)(long)cmd->info; 
+    *(BOOL *)(pConfig+offset)=TRUE;
+    return NULL;
+    }
+
+static const char *set_server_bool_slot_false (cmd_parms *cmd,char *struct_ptr)
+    {
+  /* This one should be pretty generic... */
+    char *pConfig=ap_get_module_config(cmd->server->module_config,
+				       &apache_ssl_module);
+  
+    int offset=(int)(long)cmd->info; 
+    *(BOOL *)(pConfig+offset)=FALSE;
+    return NULL;
+    }
+
+command_rec ssl_cmds[]=
+    {
+    { "SSLDisable",set_server_bool_slot,
+      (void *)XtOffsetOf(SSLConfigRec,bDisabled),RSRC_CONF,NO_ARGS,
+      "Disable SSL" },
+    { "SSLEnable",set_server_bool_slot_false,
+      (void *)XtOffsetOf(SSLConfigRec,bDisabled),RSRC_CONF,NO_ARGS,
+      "Enable SSL" },
+    { "SSLCertificateFile",set_server_string_slot,
+      (void *)XtOffsetOf(SSLConfigRec,szCertificateFile),RSRC_CONF,TAKE1,
+      "PEM certificate file" },
+    { "SSLCertificateKeyFile",set_server_string_slot,
+      (void *)XtOffsetOf(SSLConfigRec,szKeyFile),RSRC_CONF,TAKE1,
+      "Certificate private key file (assumed to be SSLCertificateFile if absent)" },
+    { "SSLCACertificatePath",set_server_string_slot,
+      (void *)XtOffsetOf(SSLConfigRec,szCACertificatePath),RSRC_CONF,TAKE1,
+      "CA Certificate path (taken from SSL_CERT_DIR if absent)" },
+    { "SSLCACertificateFile",set_server_string_slot,
+      (void *)XtOffsetOf(SSLConfigRec,szCACertificateFile),RSRC_CONF,TAKE1,
+      "CA Certificate file (taken from SSL_CERT_FILE if absent)" },
+    { "SSLVerifyDepth",set_server_int_slot,
+      (void *)XtOffsetOf(SSLConfigRec,nVerifyDepth),RSRC_CONF,TAKE1,
+      "Verify depth (default 0)" },
+    { "SSLVerifyClient",set_server_int_slot,
+      (void *)XtOffsetOf(SSLConfigRec,nVerifyClient),RSRC_CONF,TAKE1,
+      "Verify client (0=no,1=optional,2=required" },
+    { "SSLFakeBasicAuth",set_server_bool_slot,
+      (void *)XtOffsetOf(SSLConfigRec,bFakeBasicAuth),RSRC_CONF,NO_ARGS,
+      "Translate client X509 into a user name" },
+    { "SSLRequiredCiphers",set_server_string_slot,
+      (void *)XtOffsetOf(SSLConfigRec,szReqCiphers),RSRC_CONF,TAKE1,
+      "Colon-delimited list of required ciphers" },
+    /* Per Directory */
+    { "SSLRequireCipher",SSLRequireCipher,NULL,OR_FILEINFO,ITERATE,
+      "Add a cipher to the per directory list of required ciphers" },
+    { "SSLBanCipher",SSLBanCipher,NULL,OR_FILEINFO,ITERATE,
+      "Add a cipher to the per directory list of banned ciphers" },
+    { "SSLRequireSSL",SSLRequireSSL,NULL,OR_FILEINFO,NO_ARGS,
+      "Require SSL (use to protect directories that should not be served unencrypted)" },
+#if APACHE_SSL_EXPORT_CERTS
+    { "SSLExportClientCertificates",SSLExportCerts,NULL,OR_FILEINFO,NO_ARGS,
+      "Export client certificates and their certificate chain" },
+#endif
+#if CACHE_SESSIONS
+    { "SSLCacheServerPort",SSLCacheServerPort,NULL,RSRC_CONF,TAKE1,
+      "Cache server TCP port number" },
+    { "SSLSessionCacheTimeout",set_server_int_slot,
+      (void *)XtOffsetOf(SSLConfigRec,tSessionCacheTimeout),RSRC_CONF,TAKE1,
+      "Lifetime in seconds of session cache objects" },
+    { "SSLCacheServerPath",SSLCacheServerPath,NULL,RSRC_CONF,TAKE1,
+      "Path of the global cache server" },
+#endif
+    { "SSLLogFile",SSLLogFile,NULL,RSRC_CONF,TAKE1,
+      "Defunct SSL log file - DO NOT USE" },
+    { NULL },
+    };
+
 module MODULE_VAR_EXPORT apache_ssl_module =
     {
     STANDARD_MODULE_STUFF,
