@@ -65,16 +65,16 @@
 #include "gcache.h"
 #include "httpd.h"
 
-#define DEBUG		1
+#define DEBUG		0
 
 #if DEBUG
 
 void Dump(const char *szName,uchar *p,int n)
     {
-    printf("%s(",szName);
+    fprintf(stderr,"%s(",szName);
     for( ; n-- > 0 ; ++p)
-	printf("%02x",*p);
-    printf(")\n");
+	fprintf(stderr,"%02x",*p);
+    fprintf(stderr,")\n");
     }
 
 #else
@@ -131,7 +131,7 @@ void HandleGet(int nFD)
 	{
 	WriteThing(nFD,NULL,0);
 #if DEBUG
-	printf("Miss!\n");
+	fprintf(stderr,"Miss!\n");
 #endif
 	}
     else
